package CoPS.agents;

//
//  CoPSProblemSolver.java
//  CoPS
//
//  Created by Odd Erik  Gundersen on Tue May 11 2004.
//  Copyright (c) 2004 NTNU. All rights reserved.
//
// Changed by Kari 31.01.06 - Moved the behaviour into it's own class
// Changed by Kari 10.04.06 - Changed to handle several services and several conversations at the same time

//Jade classes
import jade.proto.ContractNetResponder;
import jade.core.Agent;
import jade.core.behaviours.SequentialBehaviour;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.ACLMessage;
import jade.content.lang.Codec;
import jade.content.onto.Ontology;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPAAgentManagement.RefuseException;


/**
 * This class implements the (abstract) Problem Solver agent of the CoPS framework.
 * Modified by Kari Roessland.
 */
public abstract class CoPSProblemSolver extends CoPSAgent{

	private ACLMessage currentRequest = null;
	private Ontology ontologies[];
	private Codec languages[];
	//added by Kari Roessland
	private boolean partOfTeam = false;
	
	/**
	 * The setup method, set the initial properties of the CoPSProblemSolver- agent.
	 * Modfied by Kari Roessland.
	 */
	public void setup(){
		//Initializing the language
		languages = initializeLanguages();

		for(int i = 0; i < languages.length; i++){
			getContentManager().registerLanguage(languages[i]);
		}
				
		//Initializing ontologies
		ontologies = initializeOntologies();
		
		for(int i = 0; i < ontologies.length; i++){
			getContentManager().registerOntology(ontologies[i]);
		}
		
		//Add Behaviours to this agent
		SequentialBehaviour sb = new SequentialBehaviour(this);
		sb.addSubBehaviour(new DFRegisteringBehaviour(this, this.setServiceDescription()));		
		sb.addSubBehaviour(new CheckCfpsBehaviour(this));
		this.addBehaviour(sb);	
	}
	
	protected Codec[] getLanguages(){
		return languages;
	}
	
	/**
		Method checking whether the used language is one that is known to the agent or not.
		@param message The message to check whether use a language known to the agent.
		@return boolean Returns true if the language is known to the agent and false otherwise.
	 */
	protected boolean knownLanguage(ACLMessage message){
		Codec langs[] = getLanguages();
		String messageLangName  = message.getLanguage();
		
		//If no language is specified in the message, return false.
		if(messageLangName == null) return false;
		
		//Iterate through all known languages to check whether to one used in the message is known
		for(int i = 0; i < langs.length; i++){
			if(messageLangName.equals(langs[i].getName())) return true;
		}
		return false;		
	}
	
	/**
		Method for initializing which ontologies this agent knows of.
		@return Ontology[]
	 */
	protected abstract Ontology[] initializeOntologies();
	
	/**
		Method for initializing the language this agent know.
		@return Codec The language
	 */
	protected abstract Codec[] initializeLanguages();
	
	/**
		Method checking whether the used ontology is one that is known to the agent or not.
		@param message The message to check whether use a ontology known to the agent.
		@return boolean Returns true if the ontology is known to the agent and false otherwise.
	 */
	protected boolean knownOntology(ACLMessage message){
		for(int i = 0; i < ontologies.length; i++){
			if(ontologies[i].getName().equals(message.getOntology())) return true;
		}
		
		return false;
	}
	
	/**
		Method for deceiding whether to make a proposal or not.
		@param request The request which asks for a proposal
		@return boolean Returns true if a proposal is being made
	 */
	protected abstract boolean deceideWhetherToMakeProposal(ACLMessage request);
	
	/**
		Method which will make a proposal to the given request.
		@param request The request which the agent is to make a proposal for.
		@return String A text string which is being put inside the content slot of a proposal message.
	 */
	protected abstract String makeProposal(ACLMessage request);

	/**
		Method for setting the current problem which this problem solving agent is about to solve.
		@param problem The problem described as a string.
	 */
	protected void setCurrentRequest(ACLMessage request){
		currentRequest = request;
	}
	
	/**
		Method for getting the problem this agent has agreed to solve.
		@return String The current problem being solved.
	 */
	protected ACLMessage getCurrentRequest(){
		return this.currentRequest;
	}
	
	
	/**
	 * The method getPartOfTeam returns true if the agent is part of a team
	 * and false if it is not. Added by Kari Roessland.
	 * @return boolean Whether the agent is part of a team (true) or not (false).
	 */
	protected boolean getPartOfTeam(){
		return partOfTeam;
	}
	
	/**
	 * The method setPartOfTeam is called when the agent joins a team (set to true)
	 * and when the agent is finished joining a team (set to false).
	 * NB! An agent can only join one team at a time. Added by Kari Roessland.
	 * @param boolean The value is true if the agent joins a team and false if it quits a team.
	 */
	protected void setPartOfTeam(boolean pot){
		partOfTeam = pot;
	}
	
	/**
		Method specifying how to perform the action assigned
		@param request The request specifying the action requested
		@return ACLMessage The message specifying the solution
	 */
	protected abstract ACLMessage performAction(ACLMessage request);
}
