//
//  PlanLibrary.java
//  Blackboard
//
//  Created by Kari on 01.03.06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//
package teamSpace.structure;

//java
import java.util.ArrayList;
import java.lang.reflect.Array;

//tmst
import TMST.representation.TMST; 
import TMST.representation.Task;
import TMST.representation.Action;
import TMST.representation.ProblemSolvingMethod;
import TMST.representation.ControlInformation;
import TMST.Exceptions.AlreadyHasValueException;
import TMST.representation.Input;
import TMST.representation.Goal;

//jcreek
import jcreek.representation.EntityTypeNotMatchedException;

import CoPS.FileWriter;

/**
 * This class represents the plan of the problem solving process, and serves as an interface to the TMST.
 * It has different methods to extract information from the plan.
 * 
 * @author Kari Roessland
 */
public class PlanLibrary {
	private TMST tmst;
	private String composer;
	
	private Task initialTask;
	private FileWriter fw = null;
	
	/**
	 * The constructor
	 * @param tmst The TMST (task structure tree) that has problem solving information (the plan)
	 * @param composer The name-adress of the requester of the problem
	 */
	public PlanLibrary(TMST tmst, String composer){ 
		this.tmst = tmst;
		this.composer = composer;
	}
	
	/**
	 * This method sets the FileWriter that handles printing the output of the problem solving steps to a file.
	 * @param fileWriter Reference to the FileWriter used by all agents working on the problem of one certain TR
	 */	
	public void setFileWriter(FileWriter fileWriter){
		fw = fileWriter;
	}
	
	
	/**
	 * The method returns the name of the initial task
	 * @return A string with the name of the initial task
	 **/
	public String getInitialTask(){
		if(initialTask == null){
			initialTask = tmst.getInitialProblem();
		}
		return initialTask.getName();
	}
	
	
	
	
	/**
	 * The method returns the PSM that is coupled to a task, and is part of the solution.
	 * There is only ONE PSM per task that is part of the solution.
	 * @param taskName - The name of the task which it's PSM should be returned
	 * @return The name of the PSM coupled to the task or an empty string if it is not found
	 **/
	public String getPSM(String taskName){
		String psm = "";
		Task task = null;
		ProblemSolvingMethod psms[] = null;
		
		try{
			task = new Task(tmst.getEntity(taskName));
			psms = task.getProblemSolvingMethods();
			
			if(psms == null){
				return psm;
			}
			
			for(int i = 0; i < psms.length; i++){
				if(psms[i].isSolutionPart()){
					psm = psms[i].getName();
					break;
				}
			}
			
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		
		return psm;
	}

	/**
	 * The method returns the action that is coupled to a task, and is part of the solution.
	 * There is only ONE action per task that is part of the solution.
	 * @param taskName - The name of the task
	 * @return The name of the action coupled to the task or an empty string if it is not found
	 **/	
	public String getAction(String taskName){
		String action = "";
		Task task = null;
		Action actions[] = null;
		
		try{
			task = new Task(tmst.getEntity(taskName));
			actions = task.getActions();
			
			if(actions == null){
				return action;
			}
			
			for(int i = 0; i < actions.length; i++){
				if(actions[i].isSolutionPart()){
					action = actions[i].getName();
				}
			}
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		
		return action;
	}
		
	
	/**
	 * The method returns an array with the tasks connected to a psm. If the psm
	 * is part of the solution space, all of subtasks are too.
	 * @param psmName - The name of the problem solving method
	 * @return An array of task names, returns null if no tasks are found
	 **/
	public String[] getTasks(String psmName){
		String taskNames[] = null;
		ProblemSolvingMethod psm = null;
		Task tasks[] = null;
		try{
			psm = new ProblemSolvingMethod(tmst.getEntity(psmName));
			tasks = psm.getSubtasks();
			if (tasks == null){
				return taskNames;
			}
			taskNames = new String[tasks.length];
			
			for(int i = 0; i<tasks.length; i++){
				taskNames[i] = tasks[i].getName();
			}
		
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
	
		return taskNames;
	}
	
	
	/**
	 * The method return the name of the input of an action. The method still only 
	 * handles that an action has one input.
	 * @param actionName - The name of the action
	 * @return The name of the input
	 **/
	public String getActionInput(String actionName){
		Action action = null;
		String actionInput = "";
		ArrayList inputs = new ArrayList();
		
		try{
			action = new Action(tmst.getEntity(actionName));
			if(action == null){
				return actionInput;
			}
			inputs = action.getInputsString();
			if(inputs.size() > 0){
				actionInput = (String)inputs.get(0);
			}
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		
		return actionInput;
	}

	/**
	 * The method return the name of the output of an action
	 * @param actionName - The name of the action
	 * @return The name of the output
	 **/	
	public String getActionOutput(String actionName){
		Action action = null;
		String actionOutput = "";
		
		try{
			action = new Action(tmst.getEntity(actionName));
			if(action == null){
				return actionOutput;
			}
			actionOutput = action.getOutputString();
		
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		return actionOutput;
	}
	
	/**
	 * The method return the name of the input of a task. The method still only 
	 * handles that a task has one input.
	 * @param taskName - The name of the task
	 * @return The name of the input
	 **/
	public String getTaskInput(String taskName){
		Task task = null;
		String taskInput = "";
		ArrayList inputs = new ArrayList();
		
		try{
			task = new Task(tmst.getEntity(taskName));
			if(task == null){
				return taskInput;
			}
			inputs = task.getInputsString();
			if(inputs.size() > 0){
				taskInput = (String)inputs.get(0);
			}			
			
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		
		return taskInput;
	}
	
	/**
	 * The method return the name of the output of a task
	 * @param taskName - The name of the task
	 * @return The name of the output
	 **/		
	public String getTaskOutput(String taskName){	
		Task task = null;
		String taskOutput = "";
		
		try{
			task = new Task(tmst.getEntity(taskName));
			if(task == null){
				return taskOutput;
			}
			taskOutput = task.getOutputString();
		
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		return taskOutput;	
	}

	/**
	 * The method return the name of the input of psm. The method still only 
	 * handles that a psm has one input.
	 * @param psmName - The name of the psm
	 * @return The name of the input
	 **/	
	public String getPSMInput(String psmName){
		String psmInput = "";
		ArrayList inputs = new ArrayList();
		ProblemSolvingMethod psm = null;
		
		try{
			psm = new ProblemSolvingMethod(tmst.getEntity(psmName));
			if(psm == null){
				return psmInput;
			}
			inputs = psm.getInputsString();
			if(inputs.size() > 0){
				psmInput = (String)inputs.get(0);
			}			
			
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		
		return psmInput;
	}

	/**
	 * The method return the name of the output of a psm (problem solving method)
	 * @param psmName - The name of the psm
	 * @return The name of the output
	 **/		
	public String getPSMOutput(String psmName){
		String psmOutput = "";
		ProblemSolvingMethod psm = null;
		String taskOutput = "";
		
		try{
			psm = new ProblemSolvingMethod(tmst.getEntity(psmName));
			if(psm == null){
				return psmOutput;
			}
			psmOutput = psm.getOutputString();
		
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		return psmOutput;
	
	}
	
	/**
	 * The method returns a list of strings describing IO-dependecies related to a psm.
	 * The IO-dependencies are between the psm's subtasks.
	 * @param psmName - The name of the psm
	 * @return An ArrayList with IO-dependencies
	 */
	public ArrayList getIODependencies(String psmName){
		ProblemSolvingMethod psm = null;
		ArrayList ioDepRet = new ArrayList();
		ArrayList psmCI = new ArrayList();
		ArrayList ioDep = new ArrayList();
		ControlInformation ci = null;
		String ioDepString = "";
		
		
		try{
			psm = new ProblemSolvingMethod(tmst.getEntity(psmName));
			if(psm == null){
				return ioDepRet;
			}
			psmCI = psm.getControlInformation();
			for(int i = 0; i< psmCI.size(); i++){
				ci = (ControlInformation)psmCI.get(i);
				if(ci == null){
					return ioDepRet;
				}
				ioDep = ci.getIODependencies();
				for(int j = 0; j<ioDep.size(); j++){
					ioDepRet.add(ioDep.get(j));
				}
			}
			
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		return ioDepRet;
	}
	
	/**
	 * The method returns the executor (agentname and adress) of an action
	 * @param actionName - The name of the action
	 * @return The executor of the specified action
	 **/
	public String getExecutor(String actionName){
		String executorName = "";
		Action action = null;
		
		try{
			action = new Action(tmst.getEntity(actionName));
			if(action == null){
				return executorName;
			}
			executorName = action.getExecutor().getName();
			
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}catch(AlreadyHasValueException e){
			e.printStackTrace();
		}
		
		return executorName;
	}
	
	/**
	 * The method returns the action-type of an action
	 * @param actionName - The name of the action
	 * @return The name of the action-type
	 */
	public String getActionType(String actionName){
		String actionType = "";
		Action action = null;
		
		try{
			action = new Action(tmst.getEntity(actionName));
			if(action == null){
				return actionType;
			}
			actionType = action.getType().getName();
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return actionType;
	
	}
	
	/**
	 * The method returns the composer of this problem, which is the agent, requesting the team space.
	 * @return The decomposer name and adress as a string.
	 */
	public String getComposer(){
		return composer;
	
	}
	
	/**
	 * The method returns the inputvalue of the specified input
	 * @param inputName - The specified input
	 * @return The input value as a String
	 */
	public String getInputValue(String inputName){
		String inputValue = "";
		Input input = null;
		try{
			input = new Input(tmst.getEntity(inputName));
			if(input ==null){
				return inputValue;
			}
			inputValue = input.getValue().getName();
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}catch(AlreadyHasValueException e){
			e.printStackTrace();
		}
		return inputValue;
	}
	
	/**
	 * The method returns the outputvalue of the specified output
	 * @param outputName - The specified output
	 * @return The output value as a String
	 */
	public String getOutputValue(String outputName){
		String outputValue = "";
		return outputValue;
	}
	
	/**
	 * The method returns goal of a specified task. This method only handles that a task has a single goal.
	 * @param taskName - The specified task
	 * @return The goal as a String
	 */
	public String getGoal(String taskName){
		String goalName = "";
		Task task = null;
		Goal goals[] = null;
		try{
			task = new Task(tmst.getEntity(taskName));
			if(task == null){
				return goalName;
			}
			goals = task.getGoals();
			goalName = goals[0].getName();
			
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}catch(Exception e){
			e.printStackTrace();
		}
		return goalName;
	}
	
	/**
	 * The method returns the initial goal as a string.
	 * @return The initial goal
	 */
	public String getInitialGoal(){
		return getGoal(getInitialTask());
	}
	
	/**
	 * The method returns the initial input as a string.
	 * @return The initial input
	 */
	public String getInitialInput(){
		return tmst.getInitialInput();
	}
	
	/**
	 * The method returns the teamID stored in the TMST as an int. This is the
	 * same teamID that is stored in the TeamSpace (structure) - class.
	 * @return The teamID as an int
	 */
	public int getTeamID(){
		return tmst.getTeamID();
	}
	
}
