//
//  ResultLibrary.java
//  Blackboard
//
//  Created by Kari on 01.03.06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//
package teamSpace.structure;
//java
import java.util.Iterator;
import java.util.ArrayList;

//jess
import jess.*;

import CoPS.FileWriter;

/**
 * This class represents the results and methods for extracting information about them.
 * The results are actually kept by the <code>Rete</code> rule engine. So this class also
 * serves as an interface to this rule engine.
 *
 * @author Kari Roessland
 */
public class ResultLibrary {
	private Rete ruleEngine;
	private boolean initialize = false;
	private ArrayList executedActions = new ArrayList();
	private ArrayList composedPSMs = new ArrayList();
	private FileWriter fw = null;

	/**
	 * The constructor
	 * @param ruleEngine The rule engine
	 */
	public ResultLibrary(Rete ruleEngine){
		this.ruleEngine = ruleEngine;
	}

	/**
	 * This method sets the FileWriter that handles printing the output of the problem solving steps to a file.
	 * @param fileWriter Reference to the FileWriter used by all agents working on the problem of one certain TR
	 */	
	public void setFileWriter(FileWriter fileWriter){
		fw = fileWriter;
	}
	
	/**
	 * This method add a construct to the rule engine
	 * @param construct The construct to be added
	 */
	public void add(String construct) throws JessException{
		if(initialize){
			fw.write("        Add template: "+construct);
		}
		ruleEngine.executeCommand(construct);
	}
	
	/**
	 * This method adds processed actions to the rule engine.
	 * @param actions The actions to be added, listed as {@link TSAction} objects
	 * @return boolean Returns <code>true</code> if all actions are successfully added and <code>false</code> if not
	 */
	public boolean addProcessedActions(ArrayList actions){
		boolean ret = true;
		String actionName, output, ass = "";
		TSAction action = null;
		
		for(int i = 0; i<actions.size();i++){
			action = (TSAction)actions.get(i);
			actionName = action.getName();
			output = action.getOutput();
			
			if(actionName.compareTo("") != 0 && output.compareTo("") != 0){
				try{
					ass = "(assert (executed (action "+actionName+") (output \""+output+"\")))";
					add(ass);
					executedActions.add(action);
				}catch(JessException e){
					e.printStackTrace();
					ret = false;
				}
			}
		}

		return ret;
	}

	/**
	 * This method adds composed PSMs to the rule engine.
	 * @param psms The PSMs to be added, listed as {@link TSPSM} objects
	 * @return boolean Returns <code>true</code> if all PSMs are successfully added and <code>false</code> if not
	 */	
	public boolean addProcessedPSMs(ArrayList psms){
		boolean ret = true;
		String psmName, output, ass = "";
		TSPSM psm = null;
		
		for(int i = 0; i<psms.size(); i++){
			psm = (TSPSM)psms.get(i);
			psmName = psm.getName();
			output = psm.getOutput();
			
			if(psmName.compareTo("") !=0 && output.compareTo("") !=0){
				try{
					ass = "(assert (composed (psm "+psmName+") (output \""+output+"\")))";
					add(ass);
					composedPSMs.add(psm);
				}catch(JessException e){
					e.printStackTrace();
					ret = false;
				}
			}
		}
		return ret;
	}
	
	
	/**
	 * This method return a list of the executed actions.
	 * @return The list containing {@link TSAction} objects
	 */
	public ArrayList getExecutedActions(){
		return executedActions;
	}
	
	/**
	 * This method returns a list of the composed PSMs.
	 * @return The list containing {@link TSPSM} objects
	 */
	public ArrayList getComposedPSMs(){
		return composedPSMs;
	}
	
	/**
	 * This method return a list of the actions that are ready to be executed.
	 * @return The list containing {@link TSAction} objects
	 */	
	public ArrayList getActionsReadyToExecute(){
		ArrayList actions = new ArrayList();

		Iterator facts = ruleEngine.listFacts();
		Fact fact = null;
		Value value = null;
		String actionName = "";
		String actionType = "";
		String input = "";
		String executor = "";
	
		while (facts.hasNext()){
			try{
				fact = (Fact)(facts.next());
									
				if(fact.getName().compareTo("MAIN::ready_to_execute") == 0){
					value = fact.getSlotValue("action");
					actionName = value.toString();
					value = fact.getSlotValue("action_type");
					actionType = value.toString();
					value = fact.getSlotValue("input");
					input = value.toString();
					value = fact.getSlotValue("executor");
					executor = value.toString();
					
					actions.add(new TSAction(actionName, input, executor, actionType));
				}
			}catch(JessException e){
				e.printStackTrace();
			}			
		}
		
		return actions;
	}

	/**
	 * This method returns a list of PSMs that are ready to be composed.
	 * @return The list containing {@link TSPSM} objects
	 */	
	public ArrayList getPSMsReadyToCompose(){
		ArrayList psms = new ArrayList();
		Iterator facts = ruleEngine.listFacts();
		Fact fact = null;
		Value value = null;
		String name, input, executor = "";
		ArrayList partialResults = new ArrayList();
		ValueVector vector = null;
				
		while(facts.hasNext()){
			try{
				fact = (Fact)(facts.next());
				if(fact.getName().compareTo("MAIN::ready_to_compose") == 0){
					value = fact.getSlotValue("psm");
					name = value.toString();
					value = fact.getSlotValue("input");
					input = value.toString();
					value = fact.getSlotValue("executor");
					executor = value.toString();
					value = fact.getSlotValue("partial_results");
					vector = value.listValue(ruleEngine.getGlobalContext());
					
					
					for(int i = 0; i<vector.size(); i++){
						value = (Value)vector.get(i);
						
						partialResults.add(value.toString());
					}
					
					psms.add(new TSPSM(name, input, executor, partialResults));

				}
				
			}catch(JessException e){
				e.printStackTrace();
			}
		}
		return psms;
	}
	
	/**
	 * The method initializes the result library by adding templates for how constructs (facts) should
	 * be represented in the rule base.
	 * @return <code>true</code> if the templates are successfully added and <code>false</code> if not
	 */
	public boolean initialize(){
		initialize = true;
		
		try{
			//the templates should be instantiated from the ontologies (domain knowledge and CoPS ontology)
			//CoPS ontology
			add("(deftemplate action (slot name) (slot inputname) (slot input) (slot outputname) (slot output))");
			
			add("(deftemplate taskinput (slot name) (slot value))");
			add("(deftemplate taskoutput (slot name) (slot value))");
			
			add("(deftemplate actioninput (slot name) (slot value))");
			add("(deftemplate actionoutput (slot name) (slot value))");

			add("(deftemplate psminput (slot name) (slot value))");
			add("(deftemplate psmoutput (slot name) (slot value))");
			
			add("(deftemplate ready_to_execute (slot action) (slot action_type) (slot input) (slot executor))");
			add("(deftemplate ready_to_compose (slot psm) (slot input) (multislot partial_results) (slot executor))");
			
			add("(deftemplate executed (slot action) (slot output))");
			add("(deftemplate composed (slot psm) (slot output))");
			add("(deftemplate achieved (slot goal) (slot output))");
			
					
		}catch(JessException e){
			e.printStackTrace();
			initialize = false;
			//The initialization did not succeed
			return false;
		}
		initialize = false;
		//All went OK
		return true;
	}
	

}
