//
//  RuleBase.java
//  Blackboard
//
//  Created by Kari on 01.03.06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//
package teamSpace.structure;
import java.util.ArrayList;
import jess.*;
import CoPS.FileWriter;

/**
 * This class reprsents the Rule Base of the TEAM SPACE structure and is an interface to the rules
 * in the rule engine.
 * 
 * @author Kari Roessland
 */
public class RuleBase {

	private Rete ruleEngine;
	private int ruleCount = 0;
	PlanLibrary planLib;
	FileWriter fw = null;

	/**
	 * The constructor.
	 * @param ruleEngine The rule engine
	 */
	public RuleBase(Rete ruleEngine){
		this.ruleEngine = ruleEngine;
	}

	/**
	 * This method sets the FileWriter that handles printing the output of the problem solving steps to a file.
	 * @param fileWriter Reference to the FileWriter used by all agents working on the problem of one certain TR
	 */	
	public void setFileWriter(FileWriter fileWriter){
		fw = fileWriter;
	}

	/**
	 * This method add a construct to the rule engine
	 * @param construct The construct to be added
	 */	
	public void add(String construct) throws JessException{
		String construct1, construct2;
		int index = construct.indexOf("=>");
		construct1 = construct.substring(0, index+2);
		construct2 = construct.substring(index+2, construct.length());
	
		fw.write("        Add rule: "+construct1+"\n                 "+construct2);
		
		
		ruleEngine.executeCommand(construct);
	}

	/**
	 * This method returns the next available rule-number.
	 * @return The rule-number as an <code>int</code>
	 */
	private int getRuleNumber(){
		ruleCount++;
		return ruleCount;
	}
	
	/**
	 * This method starts the initialization process of the rule base
	 * @param planLib A link to the plan library
	 */
	public boolean initialize(PlanLibrary planLib){
		boolean everythingOK = false;
		
		this.planLib = planLib;
		String initialTask = planLib.getInitialTask();
	
		try{
			everythingOK = initializeTask(initialTask);
			//add initial input (to the result-library)
			add("(assert (taskinput (name "+ planLib.getTaskInput(initialTask)+") (value \""+planLib.getInitialInput()+"\")))");
			
		}catch(JessException e){
			e.printStackTrace();
			everythingOK = false;
		}
		
		return everythingOK;
	 
	}
	
	
	/**
	 * The method translate information about a task from the tmst into rules used by
	 * the rule-engine part of the team-space.
	 * @param taskName - The name of the task
	 * @return Returns <code>true</code> if everything went ok and <code>false</code> if not
	 */	
	public boolean initializeTask(String taskName) throws JessException{
		
		String psmName = "";
		String actionName = "";
		String goalName = "";
		boolean everythingOK = false;
		String rule11 = "";
		
		//Pretty-print
		fw.write("\n    Task: "+taskName);
		
		psmName = planLib.getPSM(taskName);
		actionName = planLib.getAction(taskName);
		goalName = planLib.getGoal(taskName);
		
		//Ruletype 11: When task x has an output value, it means that the goal y of this task is reached.
		//Rule11:
		//	(defrule rule-__
		//		(taskoutput (name xoutput) (value ?out))
		//		(test (neq ?out nil))
		//	=>
		//		(assert (achieved (goal yname) (output ?out))))
		//
		rule11 = "(defrule rule-"+ getRuleNumber() + " "+
			"(taskoutput (name "+planLib.getTaskOutput(taskName)+") (value ?out)) "+ 
			"(test (neq ?out nil)) "+
			"=> (assert (achieved (goal "+goalName+") (output ?out))))";
		add(rule11);
		
		
		if (psmName.compareTo("") == 0){
			if (actionName.compareTo("") == 0){
				return everythingOK;
			}else{

				String rule1, rule2 = "";
				
				//we have got a task which is linked to an action
				//Ruletype 1: When action x has an output value, this value will be copied to task y output value. Action x performs task y. 
				//Rule1: 
				//	(defrule rule-__
				//		(actionoutput (name xoutput) (value ?out))
				//		(test (neq ?out nil))
				//	=>
				//		(assert (taskoutput (name youtput) (value ?out))))
				//
					
				rule1 = "(defrule rule-"+ getRuleNumber() + " "+
					"(actionoutput (name "+planLib.getActionOutput(actionName)+") (value ?out)) "+ 
					"(test (neq ?out nil)) "+
					"=> (assert (taskoutput (name "+planLib.getTaskOutput(taskName)+") (value ?out))))";
				add(rule1);
				
			
				//Ruletype 2: When task x has an input value, this value will be copied to action y input value. Action y performs task x.
				//Rule2:
				//	(defrule rule-__
				//		(taskinput (name xinput) (value ?in))
				//		(test (neq ?in nil))
				//	=>
				//		(assert (actioninput (name yinput) (value ?in))))				
			
				rule2 = "(defrule rule-"+getRuleNumber() + " "+
					"(taskinput (name "+planLib.getTaskInput(taskName)+") (value ?in)) "+
					"(test (neq ?in nil)) "+
					"=> (assert (actioninput (name "+planLib.getActionInput(actionName)+") (value ?in))))";
				add(rule2);
				
																
				everythingOK = initializeAction(actionName);
			}
		}else{
			String rule6, rule7, rule10 = "";
			
			//we have got a task which is linked to a psm
			//Ruletype 6: When psm x has an output value, this value will be copied to task y output value. psm x performs task y. 
			//Rule6: 
			//	(defrule rule-__
			//		(psmoutput (name xoutput) (value ?out))
			//		(test (neq ?out nil))
			//	=>
			//		(assert (taskoutput (name youtput) (value ?out))))
			//
					
			rule6 = "(defrule rule-"+ getRuleNumber() + " "+
				"(psmoutput (name "+planLib.getPSMOutput(psmName)+") (value ?out)) "+ 
				"(test (neq ?out nil)) "+
				"=> (assert (taskoutput (name "+planLib.getTaskOutput(taskName)+") (value ?out))))";
			add(rule6);
				
			
			//Ruletype 7: When task x has an input value, this value will be copied to psm y input value. psm y performs task x.
			//Rule7:
			//	(defrule rule-__
			//		(taskinput (name xinput) (value ?in))
			//		(test (neq ?in nil))
			//	=>
			//		(assert (psminput (name yinput) (value ?in))))				
			
			rule7 = "(defrule rule-"+getRuleNumber() + " "+
				"(taskinput (name "+planLib.getTaskInput(taskName)+") (value ?in)) "+
				"(test (neq ?in nil)) "+
				"=> (assert (psminput (name "+planLib.getPSMInput(psmName)+") (value ?in))))";
			add(rule7);

			everythingOK = initializePSM(psmName);
		}
		
		return everythingOK;
	}
	
	/**
	 * The method translate information about a psm from the tmst into rules used by
	 * the rule-engine part of the team-space.
	 * @param psmName - The name of the psm
	 * @return Returns <code>true</code> if everything went ok and <code>false</code> if not
	 */	
	public boolean initializePSM(String psmName) throws JessException {
		String tasks[] = null;
		boolean everythingOK = false;
	
		//Pretty-print
		fw.write("\n    PSM: "+psmName);
	
		tasks = planLib.getTasks(psmName);
		if(tasks == null){
			return everythingOK;
		}else{
			//Make the rules for the task i/o-dependecies stored in this psm
			//Ruletype 3: When task x has an output value, this value will be copied to task y input value. Task y depends upon task x.
			//Rule3:
			//	(defrule rule-__
			//		(taskoutput (name xoutput) (value ?out))
			//		(test (neq ?out nil))
			//	=>
			//		(assert (taskinput (name yinput) (value ?out))))
			
		
			String ioDep, input, output, task1, task2, rule5, rule8, rule9, rule10, taskName = "";
			int index1, index2 = 0;
			ArrayList allIODep = planLib.getIODependencies(psmName);
			
			for (int i = 0; i< allIODep.size(); i++){
				ioDep = (String)allIODep.get(i);
				fw.write("        I/O-dependecy: " + ioDep);
				//getting the input
				index1 = ioDep.indexOf("<");
				index2 = ioDep.indexOf(">");
				input = ioDep.substring(index1 + 1, index2);
				ioDep = ioDep.substring(index2 + 1);
				
				//getting the dependent task
				index1 = ioDep.indexOf("(");
				index2 = ioDep.indexOf(")");
				task1 = ioDep.substring(index1 + 1, index2);
				ioDep = ioDep.substring(index2 + 1);
				
				//getting the output				
				index1 = ioDep.indexOf("<");
				index2 = ioDep.indexOf(">");
				output = ioDep.substring(index1 + 1, index2);
				ioDep = ioDep.substring(index2 + 1);

				//getting the task that one depends upon
				index1 = ioDep.indexOf("(");
				index2 = ioDep.indexOf(")");
				task2 = ioDep.substring(index1 + 1, index2);
				
				rule5 = "(defrule rule-"+getRuleNumber() +
					"(taskoutput (name "+output+") (value ?out)) "+
					"(test (neq ?out nil)) "+
					"=> (assert (taskinput (name "+input+") (value ?out))))";
				add(rule5);
			}		
		

		
			//Ruletype 8: When all of psm y's subtasks (x1 - xn) has an outputvalue, these values will be copied to the fact that says 
			//that psm y is ready to compose.
			//Rule8:
			//	(defrule rule-__
			//		(taskoutput (name x1output) (value ?out1))
			//		(test (neq ?out1 nil))
			//		(taskoutput (name x2output) (value ?out2))
			//      (test (neq ?out2 nil))
			//      .....
			//	=>
			//		(assert (ready_to_compose (psm yname) (partialresults ?out1 ?out2 ...) (executor agentname))))
			//
			rule8 = "(defrule rule-"+getRuleNumber() + " ";
			
			for(int i = 0; i< tasks.length; i++){
				rule8 = rule8 + "(taskoutput (name " + planLib.getTaskOutput(tasks[i]) +") (value ?out" +i+ ")) " + "(test (neq ?out"+i+" nil)) ";
			
			}
			
			rule8 = rule8 + "=> (assert (ready_to_compose (psm " + psmName + ") (input "+planLib.getPSMInput(psmName)+") (partial_results";
			
			for(int i = 0; i<tasks.length; i++){
				rule8 = rule8 + " ?out"+i;
			}
			
			rule8 = rule8 + ") (executor "+planLib.getComposer() +"))))";
			
			add(rule8);
							
			
			//Ruletype 9: When psm x is composed, the output of this composition will be copied to psm x output value. And the
			//			  fact that says that psm x is ready to compose will be deleted.
			//Rule9:
			//		(defrule rule-__
			//		(composed (psm xname) (output ?out))
			//		(test (neq ?out nil))
			//		?fact <- (ready_to_compose (psm xname)) 
			//	=>
			//		(assert (psmoutput (name xoutput) (value ?out)))
			//		(retract ?fact))
		
				
			rule9 = "(defrule rule-"+getRuleNumber()+" "+
				"(composed (psm "+psmName +") (output ?out)) "+
				"(test (neq ?out nil)) "+
				"?fact <- (ready_to_compose (psm "+psmName +")) "+
				"=> (assert (psmoutput (name "+planLib.getPSMOutput(psmName) +") (value ?out))) "+
				"(retract ?fact))";
			add(rule9);	



			//Ruletype 10: When psm x has an input value, this value will be copied to task y input value, if the information in the
			//			  tmst states that these values should be the same.
			//Rule10:
			//	(defrule rule-__
			//		(psminput (name xinput) (value ?in))
			//		(test (neq ?in nil))
			//	=>
			//		(assert (taskinput (name yinput) (value ?in))))
				
			
			for(int i = 0; i<tasks.length; i++){
				taskName = tasks[i];
				fw.write("        Comparison: TaskInput-" + planLib.getInputValue(planLib.getTaskInput(taskName)) +" PSMInput-"+planLib.getInputValue(planLib.getPSMInput(psmName)));
				
				if(planLib.getInputValue(planLib.getPSMInput(psmName)).compareTo(planLib.getInputValue(planLib.getTaskInput(taskName))) == 0){
					rule10 = "(defrule rule-"+getRuleNumber() + " "+
						"(psminput (name "+planLib.getPSMInput(psmName)+") (value ?in)) "+
						"(test (neq ?in nil)) "+
						"=> (assert (taskinput (name "+planLib.getTaskInput(taskName)+") (value ?in))))";
					add(rule10);
				}
			}
			
			for(int i = 0; i< tasks.length; i++){
				everythingOK = initializeTask(tasks[i]);
				if(!everythingOK){
					return everythingOK;
				}
			}		
		
		}
		
		
		return everythingOK;
	}
	
	/**
	 * The method translate information about an action from the tmst into rules used by
	 * the rule-engine part of the team-space.
	 * @param actionName - The name of the action
	 * @return Returns <code>true</code> if everything went ok and <code>false</code> if not
	 */
	public boolean initializeAction(String actionName) throws JessException{
		String rule3, rule4 = "";
		boolean everythingOK = true;		
		fw.write("\n    Action: "+actionName);
		
		//Ruletype 3: When action x has an inputvalue, this value will be copied to the fact that says that action x is ready to execute.
		//Rule3:
		//	(defrule rule-__
		//		(actioninput (name xinput) (value ?in))
		//		(test (neq ?in nil))
		//	=>
		//		(assert (ready_to_execute (action xname) (input ?in) (executor agentname))))
		rule3 = "(defrule rule-"+getRuleNumber() + " "+
			"(actioninput (name "+planLib.getActionInput(actionName)+") (value ?in)) "+
			"(test (neq ?in nil)) "+
			"=> (assert (ready_to_execute (action "+actionName+") (action_type "+planLib.getActionType(actionName)+") (input ?in) (executor "+planLib.getExecutor(actionName) +"))))";
		add(rule3);
							
			
		//Ruletype 4: When action x is executed, the output of this execution will be copied to action x output value. And the
		//			  fact that says that action x is ready to execute will be deleted.
		//Rule4:
		//	(defrule rule-__
		//		(executed (action xname) (output ?out))
		//		(test (neq ?out nil))
		//		?fact <- (ready_to_execute (action xname)) 
		//	=>
		//		(assert (actionoutput (name xoutput) (value ?out)))
		//		(retract ?fact))
		
				
		rule4 = "(defrule rule-"+getRuleNumber()+" "+
			"(executed (action "+actionName +") (output ?out)) "+
			"(test (neq ?out nil)) "+
			"?fact <- (ready_to_execute (action "+actionName +")) "+
			"=> (assert (actionoutput (name "+planLib.getActionOutput(actionName) +") (value ?out))) "+
			"(retract ?fact))";
		add(rule4);	
		
		return everythingOK;
	}
	
	
	/**
	 * The method executes the rules in the ruleengine
	 **/
	public void executeRules(){
		try{
			ruleEngine.executeCommand("(run)");
		}catch(JessException e){
			e.printStackTrace();
		}
	}

}
