//
//  TeamSpace.java
//  Blackboard
//
//  Created by Kari on 01.03.06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

package teamSpace.structure;

//jess
import jess.*;

//tmst
import TMST.representation.TMST;
import TMST.representation.Action;
import TMST.representation.Task;

//java
import java.util.ArrayList;

import CoPS.FileWriter;


/**
 * This class represents the Team Space and it serves as an interface to all of the important
 * part of the Team Space
 * 
 * @author Kari Roessland
 */

public class TeamSpace {
	
	 
	private String teamID;
	private Rete ruleEngine;
	private RuleBase ruleBase;
	private GoalStack goalStack;
	private ResultLibrary resultLibrary;
	private PlanLibrary planLibrary; 
	private TMST taskStructureTree;
	private String requester;
	private FileWriter fw = null;
	
	
	private ArrayList processingActions = new ArrayList();
	private ArrayList processingPSMs = new ArrayList();

	
	/**
	 * The constructor instantiates the different parts of the Team Space.
	 * @param taskStructureTree A reference to the TMST from the kept by the CoPSTaskResponsible
	 * @param teamID The ID of the Team Space
	 * @param requester The requester (CoPSTaskResponsible) of the Team Space
	 */
	public TeamSpace(TMST taskStructureTree, String teamID, String requester){
		this.requester = requester;
		this.taskStructureTree = taskStructureTree;
		this.teamID = teamID;
		this.ruleEngine = new Rete();
		this.ruleBase = new RuleBase(ruleEngine);
		this.goalStack = new GoalStack(ruleEngine);
		this.resultLibrary = new ResultLibrary(ruleEngine);
		this.planLibrary = new PlanLibrary(taskStructureTree, requester);
	}
	
	/**
	 * The method initializes the Team Space and it's parts.
	 * @return Returns <code>true</code> if everything went ok and <code>false</code> if not
	 */
	public boolean initialize(){
		boolean success = false;
		
		
		fw = taskStructureTree.getFileWriter();
		goalStack.setFileWriter(fw);
		resultLibrary.setFileWriter(fw);
		planLibrary.setFileWriter(fw);
		ruleBase.setFileWriter(fw);
		
		//INITIALIZE THE RESULTLIBRARY
		fw.write("############ Initialize  Result Library and Goal Stack #########\n");
		success = resultLibrary.initialize();
		fw.write("\n################################################################");
		fw.write("\n");
		if(!success){return success;}
	
		
		//INITIALIZE THE RULEBASE
		fw.write("#############  Initialize the Rule Base Componet ###############");
		success = ruleBase.initialize(planLibrary);
		fw.write("\n################################################################");
		fw.write("\n");
		
		return success;
	}
	
	/**
	 * This method returns the FileWriter that handles printing the output of the problem solving steps to a file.
	 * @return The file writer - reference
	 */
	public FileWriter getFileWriter(){
		return fw;
	}
	
	/**
	 * The method returns the actions that are ready to be executed.
	 * @return The actions in an <code>ArrayList</code>
	 */
	public ArrayList getActionsReadyToExecute(){
		//first execute the rules in the RuleBase
		ruleBase.executeRules();
		//then get the actions ready to execute from the result library
		return resultLibrary.getActionsReadyToExecute();
	}

	/**
	 * The method returns the PSMs that are ready to be composed.
	 * @return The PSMs in an <code>ArrayList</code>
	 */	
	public ArrayList getPSMsReadyToCompose(){
		return resultLibrary.getPSMsReadyToCompose();
	}

	/**
	 * The method adds processed actions to the Result Library
	 * @param actions The actions to be added
	 * @return Returns <code>true</code> if everything went ok and <code>false</code> if not
	 */
	public boolean addProcessedActions(ArrayList actions){
		if(resultLibrary.addProcessedActions(actions)){
			return true;
		}else{
			return false;
		}
		
	}

	/**
	 * The method adds composed PSMs to the Result Library
	 * @param psms The PSMs to be added
	 * @return Returns <code>true</code> if everything went ok and <code>false</code> if not
	 */	
	public boolean addProcessedPSMs(ArrayList psms){
		if(resultLibrary.addProcessedPSMs(psms)){
			return true; 
		}else{
			return false;
		}
	}
	
	/**
	 * The method returns the solution to the initial problem
	 * @return The solution as a <code>String</code>
	 */
	public String getProblemSolution(){
		return goalStack.getGoalOutput(planLibrary.getInitialGoal());
	}
	
	/**
	 * The method returns the initial problem
	 * @return The initial problem as a <code>String</code>
	 */
	public String getInitialProblem(){
		return planLibrary.getInitialTask();
	}
	
	/**
	 * The method returns the teamID
	 * @return The teamID
	 */
	public String getTeamID(){
		return teamID;
	}

	/**
	 * The method returns the initial input
	 * @return The initial input as a <code>String</code>
	 */	
	public String getInitalInput(){
		return planLibrary.getInitialInput();
	}
	
	

	/**
	 * The method returns the processing actions - the actions to be executed
	 * in the current state of the problem solving process
	 * @return The actions in a list of {@link TSAction} objects
	 */	
	public ArrayList getProcessingActions(){
		return processingActions;
	}


	/**
	 * The method returns the processing PSMs - the PSMs to be composed
	 * in the current state of the problem solving process
	 * @return The PSMs in a list of {@link TSPSM} objects
	 */		
	public ArrayList getProcessingPSMs(){
		return processingPSMs;
	}

	/**
	 *	This method sets the actions that are going to be processed.
	 *	@param actions - A list of actions
	 */
	public void setProcessingActions(ArrayList actions){
		processingActions = actions;
	}
	
	
	/**
	 * This method sets the psms that are going to be composed.
	 * @param psms - A list of psms
	 */
	public void setProcessingPSMs(ArrayList psms){
		processingPSMs = psms;
	}

	
	/**
	 * The method prints the current state of the problem solving in the Team Space.
	 */
	public void printCurrentState(){
	
		fw.write("\n\n##############  CURRENT STATE OF THE TEAM SPACE ################\n");
		
		ArrayList goals = goalStack.getAchievedGoals();
		TSGoal goal = null;
		fw.write("######################  Achieved goals  ########################");
		if (goals.size() == 0){
			fw.write("    No achieved goals yet...\n");
		}else{
			for(int i = 0; i< goals.size(); i++){
				goal = (TSGoal)goals.get(i);
				fw.write("    Goal: "+goal.getName()+"\n    Output: "+goal.getOutput()+"\n");
			}
		}
		
		ArrayList executedActions = resultLibrary.getExecutedActions();
		TSAction action = null;
		fw.write("#####################  Executed actions  #######################");
		if(executedActions.size() == 0){
			fw.write("    No executed actions yet...\n");
		}else{
			for(int i = 0; i<executedActions.size(); i++){
				action = (TSAction)executedActions.get(i);
				fw.write("    Action: "+action.getName()+"\n    Output: "+action.getOutput()+"\n");
			}
		}

		ArrayList composedPSMs = resultLibrary.getComposedPSMs();
		TSPSM psm = null;		
		fw.write("######################  Composed PSMs  #########################");
		if(composedPSMs.size() == 0){
			fw.write("    No composed PSMs yet...\n");
		}else{
			for(int i = 0; i < composedPSMs.size(); i++){
				psm = (TSPSM)composedPSMs.get(i);
				fw.write("    PSM: "+psm.getName()+"\n    Output: "+psm.getOutput()+"\n");
			}
		}
		
		ArrayList actionsReadyToExecute = resultLibrary.getActionsReadyToExecute();
		action = null;
		fw.write("################## Actions ready to execute ####################");
		if(actionsReadyToExecute.size() == 0){
			fw.write("    No actions to execcute in the current state...\n");
		}else{
			for(int i = 0; i < actionsReadyToExecute.size(); i++){
				action = (TSAction)actionsReadyToExecute.get(i);
				fw.write("    Action: "+action.getName()+"\n    Input: "+action.getInput()+"\n    Executor: "+action.getExecutor()+"\n");
			}
		}
		
		ArrayList psmsReadyToCompose = resultLibrary.getPSMsReadyToCompose();
		ArrayList partialResults = new ArrayList();
		psm = null;
		fw.write("##################  PSMs ready to compose  #####################");
		if(psmsReadyToCompose.size() == 0){
			fw.write("    No PSMs to compose in the current state...\n");
		}else{
			for(int i= 0; i < psmsReadyToCompose.size(); i++){
				psm = (TSPSM)psmsReadyToCompose.get(i);
				fw.write("    PSM: "+psm.getName()+"\n    Executor: "+psm.getExecutor());
				partialResults = psm.getPartialResults();
				for(int j=0; j< partialResults.size(); j++){
					fw.write("    Partial result: "+(String)partialResults.get(j));
				}
				fw.write("\n");
			}
		}
	}

}
