//
//  TeamSpaceAgentBehaviour.java
//  Blackboard
//
//  Created by Kari on 01.03.06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//
package teamSpace.agent;

import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.FSMBehaviour;
import jade.domain.FIPANames;
import jade.lang.acl.ACLMessage;

/**
 * This class represents the final-state-machine behaviour that encapsulates
 * the core behaviour of the {@link TeamSpaceAgent}. The different states are implemented
 * as behaviours. The first state is {@link CheckForActionsBehaviour}, if the output of this state is
 * <code>SUCCESS</code> it transits to {@link HandleActionsBehaviour}, if the output is <code>NO_MORE_ACTIONS</code>
 * it transits to {@link EndProblemSolvingBehaviour}. If the output of the state {@link HandleActionsBehaviour} is
 * <code>SUCCESS</code> it transits back to the {@link CheckForActionsBehaviour}. If the output of the state 
 * {@link EndProblemSolvingBehaviour} is <code>SUCCESS</code> it transits to the state {@link PrepareResultBehaviour}.
 * If any of these states transits with a corresponding failure-value <code>CHECK_FOR_ACTIONS_FAILED</code>, <code>HANDLE_ACTIONS_FAILED</code>,
 * <code>END_PROLEM_SOLVING_FAILED</code> or <code>PREPARERESULT_BEHAVIOUR_FAILED</code>, the final state will be the {@link FailureBehaviour}.
 * Else the final state will be the state {@link PrepareResultBehaviour}.
 *
 * @author Kari Roessland
 */
public class TeamSpacePSBehaviour extends FSMBehaviour{

	private static final String CFA = "CheckForActions";
	private static final String HA = "HandleActions";
	private static final String EPS = "EndProblemSolving";
	private static final String PR = "PrepareResultBehaviour";
	private static final String F = "Failure";
	
	public static final int SUCCESS = 0;
	public static final int CHECK_FOR_ACTIONS_FAILED = 1;
	public static final int HANDLE_ACTIONS_FAILED = 2;
	public static final int END_PROBLEM_SOLVING_FAILED = 3;
	public static final int PREPARERESULT_BEHAVIOUR_FAILED = 4;
	public static final int NO_MORE_ACTIONS = 5;
	
	private String teamID = "";
	
	/** 
	 * The constructor calls the super-constructor.
	 * @param agent The owner agent of this behaviour
	 */
	public TeamSpacePSBehaviour(Agent agent, String teamID){
		super(agent);
		this.teamID = teamID;
	}
	
	/** 
	 * The method initializes the FSM-behaviour and the different transitions
	 */
	public void onStart(){
		
		ACLMessage actionRequest = new ACLMessage(ACLMessage.REQUEST);
		actionRequest.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);

		
		registerFirstState(new CheckForActionsBehaviour(myAgent, teamID), CFA);
		//for now the teamspaceagent only handle one action at the time
		registerState(new HandleActionsBehaviour(myAgent, actionRequest, teamID), HA);
		registerState(new EndProblemSolvingBehaviour(myAgent, teamID), EPS);
		registerLastState(new PrepareResultBehaviour(myAgent, this.getDataStore(), this.getParent(), teamID), PR);
		registerLastState(new FailureBehaviour(myAgent, this.getDataStore(), this.getParent(), teamID), F);
		
		
		//Register the default transitions - return of SUCCESS (0)
		registerDefaultTransition(CFA, HA);
		registerDefaultTransition(HA, CFA);
		registerDefaultTransition(EPS, PR);
		
		//Register the alternative transition - return of other values but SUCCESS (0)
		registerTransition(CFA, EPS, NO_MORE_ACTIONS);
		registerTransition(CFA, F, CHECK_FOR_ACTIONS_FAILED);
		registerTransition(HA, F, HANDLE_ACTIONS_FAILED);
		registerTransition(PR, F, PREPARERESULT_BEHAVIOUR_FAILED);
		registerTransition(EPS, F, END_PROBLEM_SOLVING_FAILED);
		
	}
	
	/**
	 * The method resets this behaviour and returns the transition from this state (behaviour).
	 * @return The transition value
	 */	
	public int onEnd(){
		this.reset();
		return 0;
	}

}
