package TMST.representation;

//
//  Assumptions.java
//  TMST
//
//  Created by Odd Erik  Gundersen on Fri Apr 02 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//

//Java
import java.util.Collection;
import java.lang.Exception;

//jCreek
import jcreek.representation.Entity;
import jcreek.representation.Relation;
import jcreek.representation.RelationType;
import jcreek.representation.EntityType;
//import jcreek.representation.cbr.IsoPodModel;
import jcreek.representation.BasicModel;
import jcreek.representation.KnowledgeModel;
import jcreek.representation.NameAlreadyExistException;
import jcreek.representation.EntityTypeNotMatchedException;

//TMST
import TMST.Exceptions.AlreadyHasValueException;

/**
	Class describing assumptions of problem solving methds. Assumptions may be classified in a specialization
	hierarchy. This classification is based on the paper Assumptions of PSMs by Benjamins and Pierret-Golbreich:
	<ul>
		<li><b>Epistemological assumptions</b> refers to the domain knowledge required by the PSM for fillen static 
			input and dynamic input. Epistemological assumptions are divided into two categories:
			<ol>
				<li><b>Availability assumptions</b> refer to availability of domain knowledge.</li>
				<li><b>Property assumptions</b> refers to properties of the domain knowledge.</li>
			</ol>
		</li>
 		<li><b>Pragmatic assumptions</b> refers to PSM requirements related to the physical context or external environmetn
			in which the system operates.</li>
 		<li><b>Teleological assumptions</b> relate to the goal to be achieved, that is, a functional spesification.</li>
	</ul>

	There is only four different classes in the implementation, namly:
	<ul>
		<li>Availability assumptions</li>
		<li>Property assumptions</li>
		<li>Pragmatic assumptions</li>
		<li>Teleological assumptions</li>
	</ul>
	Thus, the superclass of epistemological assumption has been avoided, and it has not been made explicit for 
	the sake of simplicity.
 
*/
public class Assumption extends EntityType{
	/**
		The constructor of assumptions initializes a new assumption object by setting up the properties
		and ensuring that the superclasses of assumption and the assumption type is in the model. There exist 
		four different types of assumptions, which are listed in the field summary.
		@param tmst The knowledge model this assuption is a part of.
		@param name The name of the assumption
		@param description The description, as words, of what is being assumed
		@param assType The type of assumption being instatiated
		@param value A string representing the value of this assumption
	*/
	public Assumption(TMST tmst, String name, String description, int assType, String value) throws NameAlreadyExistException, Exception {
		super((KnowledgeModel)tmst, name, description);
		System.out.println("Inne i konstruktoren til ASSUMPTION");
		
		//Declarations
		Entity assumptionType = null;
		String type=null;
		
		if(assType<0 && assType>3){
			throw new Exception("The specified assumption type does not exist. Something is terrible wrong!");
		}
		
		switch(assType){
			case AssumptionType.EPISTEMOLOGICAL_AVAILABILITY:
				type="EPISTEMOLOGICAL_AVAILABILITY";
				break;
			case AssumptionType.EPISTEMOLOGICAL_PROPERTY:
				type="EPISTEMOLOGICAL_PROPERTY";
				break;
			case AssumptionType.PRAGMATIC:
				type="PRAGMATIC";
				break;
			case AssumptionType.TELEOLOGICAL:
				type="TELELOGICAL";
				break;
			default:
				throw new Exception("The Assumption type used does not exist.");
		}

		//What happens here is that the new relation with the given name is related to _the_ assumption
		//in the model -> means that it is an assumption. In addition, a relation to the assumption type is 
		//made. This relation constitutes the fact that the given assumption being made is of the chosen type.
		//This would (in a frame based model) mean that an assumption has a slot "assumption type" with the 
		//assumption type as the value.
		try{
			//The model need an assumption object 
			if(!tmst.hasEntity("ASSUMPTION")){
				Entity goal= new Entity((KnowledgeModel)tmst, "ASSUMPTION", "Generic assumption object all assumptions is an instance of.");
				goal.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity(BasicModel.SYMBOL));
			}
			this.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity("ASSUMPTION"));
			
			//The model need an object of the assumption type this assumption has:
			if(!tmst.hasEntity(type)){
				try{
					assumptionType = new AssumptionType(tmst, type, type);
					assumptionType.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity(BasicModel.SYMBOL));
				}catch(NameAlreadyExistException e){
					System.out.println("An assumption type with the following name already exist: " +e.getMessage());
				}
			}		
			
			//Relate the value to this output			
			RelationType hasType = tmst.getRelationType("HAS_ASSUMPTION_TYPE");
			
			if(hasType==null){
				//Has to make new Relation type
				hasType = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_ASSUMPTION_TYPE", "Relation pointing to the assumption type of this assumption", 1.0);
			}
			new Relation(this, tmst.getRelationType("HAS_ASSUMPTION_TYPE"), tmst.getEntity(type));
			
			//Have to relate the value of the assumption to the assumption
			RelationType hasValue = tmst.getRelationType("HAS_VALUE");
			
			if(hasValue==null){
				//Have to make a new relation type if the type has_value do not exist
				hasValue = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_VALUE", "Relation pointing to the value of this assumption", 1.0);
			}
			
			try{
				new Value(tmst, value, value);
			}catch(NameAlreadyExistException e){
				System.out.println("ASSUMPTION: " +e.getMessage());
			}
			
			new Relation(this, hasValue, tmst.getEntity(value));
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	public Assumption(Entity ent) throws EntityTypeNotMatchedException{
		super(ent);
	}
	
	/**
		Method returning the type of an assumption
		@return AssumptionType The type od the assuption
	 */
	public AssumptionType getType() throws Exception{	
		//Find the relation pointing to a assumption type (its a "has type" relation)
		Relation rels[] = this.getLocalRelations("HAS_ASSUMPTION_TYPE");

		System.out.println("LENGTH OF HAS_TYPE ARRAY: " +rels.length);
		
		if(rels.length==0){
			throw new Exception("An assumption must have a type.\n Something is terrible wrong!");
		}
		
		//An assumption may only have one type
		if(rels.length>1){
			throw new Exception("An assumption may only have one assumption type.\n Something is terrible wrong");
		}
		
		//Return the type pf the assumption
		return new AssumptionType(rels[0].getValue());
	}
	
	/**
		Method returning the value of the assumption.
		@return Value The value of the assumption
	 */
	public Value getValue() throws AlreadyHasValueException{
		
		Value value = null;
		Relation rels[] = this.getLocalRelations("HAS_VALUE");
		
		if(rels==null){
			return value;
		}
		if(rels.length>1){
			throw new AlreadyHasValueException("An output may only have one value. Something is terrible wrong");
		}
		try{
			value = new Value(rels[0].getValue());
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
		}
		return value;		
	}
	
	
	/**
		Method which has to be overriden (from the class EntityType)
		@param value The entity to match against this entity type
		@return boolean Returns true if the entity value is of this entity type.
	*/
	public static boolean matches(Entity value){
		TMST tmst = (TMST)value.getKnowledgeModel();

		Entity ent = tmst.getEntity("ASSUMPTION");
		
		if (ent == null) return false; 
		
		Collection subclasses = ent.getAllSubclassesCollection();
		
		if(value.getName().equals("ASSUMPTION")){
			return false;
		}else if(subclasses.contains(value)){
			return true;
		}
		return false;
	}
}
