package TMST.representation;

//
//  Executor.java
//  TMST
//
//  Created by Odd Erik  Gundersen on Sat May 22 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//
//	10.04.06 Changed by Kari Roessland: Changed the method for relating an action to an executor (addAction).
//										This will allow an executor to be related to several actions without
//										any failures occuring

//Java
import java.util.Collection;

//jCreek
import jcreek.representation.Entity;
import jcreek.representation.Relation;
import jcreek.representation.RelationType;
import jcreek.representation.EntityType;
//import jcreek.representation.cbr.IsoPodModel;
import jcreek.representation.BasicModel;
import jcreek.representation.KnowledgeModel;
import jcreek.representation.NameAlreadyExistException;
import jcreek.representation.EntityTypeNotMatchedException;

//TMST
import TMST.Exceptions.AlreadyHasValueException;

/**
	Class representing an agent which has volunteered to perform (or execute) an action.
*/
public class Executor extends EntityType{


	public Executor(TMST tmst, String name, String description, Action action, String costValue, String reference) throws NameAlreadyExistException{
		super((KnowledgeModel)tmst, name, description);
		try{
			//Initialization
			Cost cost = null;
			String costName = name +"-" +action.getName() + "-cost"; 
			
			//The model need a generic EXECUTOR entity 
			if(!tmst.hasEntity("EXECUTOR")){
				Entity exec = new Entity((KnowledgeModel)tmst, "EXECUTOR", "EXECUTOR");
				exec.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity(BasicModel.SYMBOL));
			}	
			this.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity("EXECUTOR"));
			
			
			//Relate the cost to this executor
			RelationType hasValue = tmst.getRelationType("HAS_COST");
			
			if(hasValue==null){
				//Has to make new Relation type
				hasValue = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_COST", "Relation pointing to a value representing cost", 1.0);
			}

			//Make a new cost object and relate it to the executor
			cost = new Cost(tmst, costName, "executor cost", this, costValue);
			cost.setReplyReference(new ReplyReference(tmst, name +action.getName() +"-reference", "reference" ,reference));
			
			action.setExecutor(this);
			
			//The cost should be set as the cost of an action
			//If it is not the lowest cost, it will relate to the action
			//The action only relates to the cost belonging to the executor which has the lowest cost.
			action.setExecutionCost(cost);
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	public Executor(Entity ent) throws EntityTypeNotMatchedException{
		super(ent);
	}
	
	/**
		Method returning all actions this executor may perform.
		@return Action[] All actions this executor may perform
	 */
	public Action[] getActions(){
		Relation rels[] = this.getLocalRelations("HAS_EXECUTOR");
		
		if(rels==null) return null;
		
		Action actions[] = new Action[rels.length];
		
		for(int i = 0; i < rels.length; i++){
			actions[i] = (Action)rels[i].getValue();
		}
		
		return actions;
	}
	
	/**
		Method for adding an action to the executor
		@param action the action name to add
		@param the cost value as a String
		@param the reference to add to the cost entity
	 */
	public void addAction(String actionName, String costValue, String reference) throws Exception{
		String costName = "";
		Cost cost = null;
		Action action = null;
		TMST tmst = (TMST)this.getKnowledgeModel();
		Entity ent = this.getKnowledgeModel().getEntity(actionName);
		
		//Check whether there exist an action with the secified name in the tmst
		if(!Action.matches(ent)) throw new Exception("No Such Action");

		//Relate the cost to this executor
		RelationType hasValue = tmst.getRelationType("HAS_COST");
		
		if(hasValue==null){
			//Has to make new Relation type
			hasValue = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_COST", "Relation pointing to a value representing cost", 1.0);
		}
				
		//Make a new cost object and relate it to the executor
		costName = this.getName() +"-" +actionName + "-cost"; 
		cost = new Cost(tmst, costName, "executor cost", this, costValue);
		cost.setReplyReference(new ReplyReference(tmst, this.getName() + actionName +"-reference", "reference", reference));
		
		
		//Instantiate the action and set this executor as an executor of the action
		action = new Action(ent);		
		action.setExecutor(this);

		//The cost should be set as the cost of an action
		//If it is not the lowest cost, it will relate to the action
		//The action only relates to the cost belonging to the executor which has the lowest cost.
		action.setExecutionCost(cost);
	}
	
	/**
		Method returning the specified cost. If this executor has no cost related to that action, null is returned.
		@param actionName The name of the action which cost information is requested
		@return Cost The cost of the this executor performing the specified action ()
	 */
	public Cost getCost(String actionName){
		Cost cost = null;
		String costName = this.getName() + "-" +actionName +"-cost";
		
		try{
			cost = new Cost(this.getKnowledgeModel().getEntity(costName));
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
			return null;
		}
		return cost;
	}
	
	/**
		Method returning the specified cost. If this executor has no cost related to that action, null is returned.
		@param actionName The name of the action which cost information is requested
		@return Cost The cost of the this executor performing the specified action ()
	 */
	public float getCostValue(String actionName) throws AlreadyHasValueException{
		Cost cost = null;
		String costName = this.getName() + "-" +actionName +"-cost";
		
		try{
			cost = new Cost(this.getKnowledgeModel().getEntity(costName));
		}catch(EntityTypeNotMatchedException e){
			e.printStackTrace();
			return Float.POSITIVE_INFINITY;
		}
		return cost.getFloatValue();
	}
	
		
	
	/**
		Overriding EntityType's method
	 */
	public static boolean matches(Entity value){
		TMST tmst = (TMST)value.getKnowledgeModel();
		Entity ent = tmst.getEntity("EXECUTOR");
		
		if (ent == null) return false; 
		
		Collection subclasses = ent.getAllSubclassesCollection();
		
		if(value.getName().equals("EXECUTOR")){
			return false;
		}else if(subclasses.contains(value)){
			return true;
		}
		return false;
	}
}
