package TMST.representation;
//
//  ProblemSolvingMethod.java
//  TMST
//
//  Created by Odd Erik  Gundersen on Mon Feb 16 2004.
//  Copyright (c) 2004 NTNU. All rights reserved.
//

//Java
import java.util.Collection;
import java.util.ArrayList;

//jCreek
import jcreek.representation.EntityType;
//import jcreek.representation.cbr.IsoPodModel;
import jcreek.representation.BasicModel;
import jcreek.representation.Entity;
import jcreek.representation.Relation;
import jcreek.representation.RelationType;
import jcreek.representation.KnowledgeModel;
import jcreek.representation.LocalKnowledgeModel;
import jcreek.representation.NameAlreadyExistException;
import jcreek.representation.NoSuchRelationTypeException;
import jcreek.representation.NoSuchRelationException;
import jcreek.representation.EntityTypeNotMatchedException;

//TMST
import TMST.Exceptions.AlreadyHasValueException;

/**
	Class describing Problem Solving Methods. PSMs describe what to do by pointing to tasks 
	which solve a certain task which refers to the PSM. Parameters the PSM may hold:
	<ul>
		<li>Assumpotions</li>
		<li>Inputs</li>
		<li>Outputs</li>
		<li>Control Information</li>
		<li>Subtasks</li>
	</ul>
	
	Comment: (Kari Roessland) All the getMethods should return ArrayLists and not regular lists.
*/
public class ProblemSolvingMethod extends EntityType{
	
	
	/**
		Constructor for Problem Solving Methods
		@param tmst The TMST this problem solving method belongs to
		@param name The name of the problem solving method
		@param description Description of the PSM
	*/
	public ProblemSolvingMethod(TMST tmst, String name, String description) throws NameAlreadyExistException{
		super((KnowledgeModel)tmst, name, description);
		
		//The model need a generic psm to 
		if(!tmst.hasEntity("PROBLEM_SOLVING_METHOD")){
			try{
				Entity psm = new Entity((KnowledgeModel)tmst, "PROBLEM_SOLVING_METHOD", "A generic problem solving method all PSMs relate to.");
				psm.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity(BasicModel.SYMBOL));
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
		//Add a relation between the generic psm and this instance of it.
		try{
			this.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity("PROBLEM_SOLVING_METHOD"));			
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
		Cosntructor for encpsulating an entity as a psm (to be allowed to use psm methods on it)
		@param ent The entity to encapsulate.
	 */
	public ProblemSolvingMethod(Entity ent) throws EntityTypeNotMatchedException {
		super(ent);
	}

	/**
		Method for checking whether a PSM is part of the solution space or not.
		@return boolean The method returns true if part of the solution space and false if not.
	 */
	public boolean isSpacePart(){
		try{
			if(this.getLocalRelation("IS_SPACE_PART")!=null){
				return true;
			}
		}catch(NoSuchRelationTypeException e){
			//If there is no such relationtype, this task is not activated.
			//Nothing need to be done and false is returned.
		}catch(NoSuchRelationException e){
			//If there is no such relation, this task is not activated.
			//Nothing need to be done and false is returned.
		}
		return false;
	}
	
	/**
	Method for checking whether a psm is part of the solution or not.
	 @return boolean The method returns true if part of the solution and false if not.
	 */
	public boolean isSolutionPart(){
		try{
			if(this.getLocalRelation("IS_SOLUTION_PART")!=null){
				return true;
			}
		}catch(NoSuchRelationTypeException e){
			//If there is no such relationtype, this task is not activated.
			//Nothing need to be done and false is returned.
		}catch(NoSuchRelationException e){
			//If there is no such relation, this task is not activated.
			//Nothing need to be done and false is returned.
		}
		return false;
	}
	
	
	/**
		Method which checks whether all subtasks of this problem solving method are activated. If they are true will be returned, else false.
		@param boolean
	 */
	protected boolean isFulFilled(){
		Task subtasks[] = this.getSubtasks();
		
		//Iterate through the subtasks of this psm, return false if one of them is not fullfilled.
		for(int i = 0; i <subtasks.length; i++){
			if(!subtasks[i].isActivated()) return false;
		}
		
		return true;
	}
	
	
	/**
		Method for setting the cost of a PSM. Takes a float as parameter.
		@param cost The cost value as flaot which is being related to the task.
	 */
	protected void setCost(float costValue) throws AlreadyHasValueException {
		Cost cost = null;
		RelationType hasCost = getKnowledgeModel().getRelationType("HAS_COST");
		
		//Make a has cost relation type if it does not exist
		try{
			if(hasCost == null) hasCost = new RelationType(getKnowledgeModel(), getKnowledgeModel().getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_COST", "Relation pointing to an a cost entity.", 1.0);
		}catch(NameAlreadyExistException e){
			e.printStackTrace();
		}
		
		//Get all local relations of type has cost
		Relation rels[] = this.getLocalRelations("HAS_COST");
		
		//If no local relation of type has cost, relate the cost specified to this PSM
		if(rels.length == 0){
			String costName = getName() +"-cost";
			try{
				cost = new Cost((TMST)getKnowledgeModel(), costName, "psm cost", (Entity)this, Float.toString(costValue));
			}catch(NameAlreadyExistException e){
				e.printStackTrace();
			}

			//if (cost != null) new Relation(this, hasCost, cost);
		}else{
			System.out.println("THIS TASK HAS MORE THAN ONE COST RELATED TO IT:");
			for(int i = 0; i < rels.length; i++){
				System.out.println("	|-- Cost name #" +i +" " +rels[i].getValue().getName());
			}
			throw new AlreadyHasValueException("A task may only hvae one cost related to it. Something is very wrong!");
		}
	}
	
	/**
		Method returning the cost entity of this psm.
		@return Cost
	 */
	public Cost getCost() throws AlreadyHasValueException{
		Relation rels[] = this.getLocalRelations("HAS_COST");
		
		if(rels.length==1) return (Cost)rels[0].getValue();
		else throw new AlreadyHasValueException("A psm may only hvae one cost related to it. Something is very wrong!");
	}
	
	/**
	Method returning the cost entity of this psm.
	 @return Cost
	 */
	public float getCostValue() throws AlreadyHasValueException{
		Relation rels[] = this.getLocalRelations("HAS_COST");
		
		if(rels.length==1){
			return ((Cost)rels[0].getValue()).getFloatValue();
		}else{
			throw new AlreadyHasValueException("A psm may only have one cost related to it. Something is very wrong!");
		}
	}
				
	/**
		Method for setting one subtask for this PSM to point to
		@param subtask The task to point to.
	*/
	public void setSubtask(Task subtask){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasTask = tmst.getRelationType("HAS_TASK");
		
		if(hasTask==null){
			//Has to make new Relation type
			try{
				hasTask = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_TASK", "Relation pointing to an task, usually from a problem solving method.", 1.0, "HAS_SOLUTION", "The inverse relation of HAS_TASK", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		new Relation(this, hasTask, subtask);
	}

	/**
		Method for setting a set of subtasks as the tasks this PSM points to.
		@param subtasks The set of subtasks which is to be associated with this psm.
	*/
	public void setSubtasks(Task[] subtasks){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasTask = tmst.getRelationType("HAS_TASK");
		
		if(hasTask==null){
			//Has to make new Relation type
			try{
				hasTask = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_TASK", "Relation pointing to a task, usually from a problem solving method.", 1.0, "HAS_SOLUTION", "The inverse relation of HAS_TASK", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
		//for every goal in the array goals, make a relation between the goal and this task
		for(int i = 0; i<subtasks.length; i++){
			try{
				new Relation(this, hasTask, subtasks[i]);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
	}
	
	/**
		A method that returns all subtasks that this problem solving method points to
		@return Task[] The subtasks which this psm points to
	*/
	public Task[] getSubtasks(){
		Task subtasks[] = null;

		Relation rels[] = this.getLocalRelations("HAS_TASK");
		if(rels==null){
			return null;
		}
		subtasks = new Task[rels.length];
		
		for(int i=0; i<rels.length; i++){
			try{
				subtasks[i]= new Task(rels[i].getValue());
			}catch(EntityTypeNotMatchedException e){
				e.printStackTrace();
			}
		}		
		return subtasks;
	}

	/**
		Method for adding a value to the control information of this PSM
		@param value Value to add to the control information, as a String.
	 */
	public void setControlInformation(String value){
		TMST tmst = (TMST)this.getKnowledgeModel();		
		ControlInformation ci = null;
		String ciName = this.getName() +"_control_information";
		RelationType hasCI = tmst.getRelationType("HAS_CONTROL_INFORMATION");
		
		//kari - debug
		System.out.println("\n set controlinformation: "+value+ "for psm: "+this.getName()+"\n");
		
		//Check whether there exist a relation type called HAS_RELATION in the tmst, if not, make it.
		if(hasCI==null){
			//Has to make new Relation type
			try{
				hasCI = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_CONTROL_INFORMATION", "Relation pointing to control information, usually from a problem solving method.", 1.0);
			}catch(NameAlreadyExistException e){
				e.printStackTrace();
			}
		}
		
		//Relation rels[] = this.getLocalRelations("HAS_CONTROL_INFORMATION");
		//Have to make this psm a control information entity for keeping control information in, if this psm does not have it.
		if(!tmst.hasEntity(ciName)){
			try{
				new ControlInformation(tmst, ciName, "Control information related to the PSM named " +this.getName());
				new Relation(this, hasCI, tmst.getEntity(ciName), 1.0);
			}catch(NameAlreadyExistException e){
				e.printStackTrace();
			}
		}
		//ci = (ControlInformation)tmst.getEntity(ciName);
		try{
			ci = new ControlInformation(tmst.getEntity(ciName));
		}catch(Exception e){
			e.printStackTrace();
		}
		
		//When sure of this psm is related to control information, add the value.
		ci.setIODependency(value);
	}
	
	/**
		Method for adding a value to the control information of this PSM
		@param value Value to add to the control information, as a Value class instance.
	 */
	public void setControlInformation(Value value){
		TMST tmst = (TMST)this.getKnowledgeModel();		
		ControlInformation ci = null;
		RelationType hasCI = tmst.getRelationType("HAS_CONTROL_INFORMATION");
		
		//kari - debug
		System.out.println("\n set controlinformation: "+value.getName()+ "for psm: "+this.getName() +"\n");

		
		//Check whether there exist a relation type called HAS_RELATION in the tmst, if not, make it.
		if(hasCI==null){
			//Has to make new Relation type
			try{
				hasCI = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_CONTROL_INFORMATION", "Relation pointing to control information, usually from a problem solving method.", 1.0);
			}catch(NameAlreadyExistException e){
				e.printStackTrace();
			}
		}
		
		Relation rels[] = this.getLocalRelations("HAS_CONTROL_INFORMATION");
		//Have to make this psm a control information entity for keeping control information in, if this psm does not have it.
		if(rels==null){
			try{
				ci = new ControlInformation(tmst, this.getName() +"_control_information", "Control information related to the PSM named" +this.getName());
				new Relation(this, hasCI, ci, 1.0);
			}catch(NameAlreadyExistException e){
				e.printStackTrace();
			}
		}
		//When sure of this psm is related to control information, add the value.
		ci.setIODependency(value);
	}	
	
	
	/**
		Method for updating the control information of this PSM
		@param ci The control information which is to be associated with the psm.
	*/
	public void updateControlInformation(ControlInformation ci) throws Exception{
		ControlInformation retCI = null;
		Relation rels[] = this.getLocalRelations("ControlInformation");
		System.out.println("updateControlInformation: METHOD NOT FINISHED!");
		
		if(rels==null){
			System.out.println("This PSM has no ControlInformation to update.");
		}else{
			System.out.println("Yeah. Come on, just update.");
		}
	}

	/**
	 *	A method that returns all Constraints that apply for this psm
	 *	@return ArrayList The control information associated to the psm.
	 *  Changed by Kari Roessland
	 */
	public ArrayList getControlInformation(){
		
		Relation rels[] =this.getLocalRelations("HAS_CONTROL_INFORMATION");
		Entity ent = null;
		ControlInformation ci = null;
		ArrayList retCI = new ArrayList();
		
		if (rels != null){		
			for(int i=0; i<rels.length; i++){
				try{
					ent = rels[i].getValue();
					ci = new ControlInformation(ent);
					retCI.add(ci);
				
				}catch(EntityTypeNotMatchedException e){
					e.printStackTrace();
				}
			}
		}
		return retCI;
	}
	
	/**
		Method for setting an output for the psm.
		@param output The output 
	*/	
	public void setOutput(Output output){
		TMST tmst = (TMST)this.getKnowledgeModel();		
		RelationType hasOutput = tmst.getRelationType("HAS_OUTPUT");
		
		if(hasOutput==null){
			//Has to make new Relation type
			try{
				hasOutput = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_OUTPUT", "Relation pointing to an output", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		new Relation(this, hasOutput, output);
	}
	
	/**
		Method for setting a set of outputs for the psm.
		@param outs The set of output 
	*/
	public void setOutputs(Output[] outs){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasOutput = tmst.getRelationType("HAS_OUTPUT");
		
		if(hasOutput==null){
			//Has to make new Relation type
			try{
				hasOutput = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_OUTPUT", "Relation pointing to an output", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
		//for every goal in the array goals, make a relation between the goal and this task
		for(int i = 0; i<outs.length; i++){
			try{
				new Relation(this, hasOutput, outs[i]);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
	}
	
	/**
		Method which returns any output related to this psm
		@return Output Yeah, the output ...
	*/
	public Output[] getOutput(){
		Relation rels[] = this.getLocalRelations("OUTPUT");
		if(rels==null){
			return null;
		}
		
		Output output[] = new Output[rels.length];
		for(int i=0; i<rels.length; i++){
			output[i]=(Output)rels[i].getValue();
		}
		return output;
	}
	
	
	/**
	 * Method for getting the output entity as a string - this method assume that there is only one output.
	 * @author Kari Roessland
	 * @return String The name of the output entity
	 */
	public String getOutputString(){
		Relation[] rels = this.getLocalRelations("HAS_OUTPUT");
		String retOutput = "";
		Output output = null;
		if (rels != null && rels.length == 1){
			try{
				output = new Output(rels[0].getValue());
				retOutput = output.getName();
			}catch(EntityTypeNotMatchedException e){
				e.printStackTrace();
			}
		}
		return retOutput;
	}

	
	
	/**
		Method for setting input of the PSM
		@param input The input.
	*/
	public void setInput(Input input){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasInput = tmst.getRelationType("HAS_INPUT");
		
		if(hasInput==null){
			//Has to make new Relation type
			try{
				hasInput = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_INPUT", "Relation pointing to a input", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		new Relation(this, hasInput, input);
	}
	
	/**
		Method which sets the input needed for to solve the psm >> Assumptions?
		@param ins The set of inputs assigned to the PSMs
	*/
	public void setInputs(Input[] ins){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasInput = tmst.getRelationType("HAS_INPUT");
		
		if(hasInput==null){
			//Has to make new Relation type
			try{
				hasInput = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_INPUT", "Relation pointing to an input", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
		//for every goal in the array goals, make a relation between the goal and this task
		for(int i = 0; i<ins.length; i++){
			new Relation(this, hasInput, ins[i]);
		}
	}
	
	/**
		Method which returns the input this method requires
		@return Input[] The method returns an array containing all inputs associated with the psm.
	*/
	public Input[] getInput(){
		Relation rels[] = this.getLocalRelations("INPUT");
		if(rels==null){
			return null;
		}
		
		Input input[] = new Input[rels.length];
		for(int i=0; i<rels.length; i++){
			input[i]=(Input)rels[i].getValue();
		}
		return input;
	}

	/**
	 * Method for getting all input-entities as strings
	 * @author Kari Roessland
	 * @return ArrayList A list of input-entities as strings
	 */
	public ArrayList getInputsString(){
		Relation rels[] = this.getRelations("HAS_INPUT");
		ArrayList retInput = new ArrayList();
		Input input = null;
		
		if (rels != null && rels.length > 0){
			for (int i = 0; i< rels.length; i++){
				try{
					input = new Input(rels[i].getValue());
					retInput.add(input.getName());
				}catch(EntityTypeNotMatchedException e){
					e.printStackTrace();
				}
			}
		}
		return retInput;
	}
	
	/**
		Method for setting the assumptions related to the psm.
		@param ass The assumptions which apply for the psm
	*/ 
	public void setAssumption(Assumption ass){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasAss = tmst.getRelationType("HAS_ASSUMPTION");
		
		if(hasAss==null){
			//Has to make new Relation type
			try{
				hasAss = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_ASSUMPTION", "Relation pointing to an input", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}		
		new Relation(this, hasAss, ass);
	}
	
	/**
		Method for setting the assumptions related to the psm.
		@param ass The assumptions which apply for the psm
	 */ 
	public void setAssumptions(Assumption[] ass){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType hasAss = tmst.getRelationType("HAS_ASSUMPTION");
		
		if(hasAss==null){
			//Has to make new Relation type
			try{
				hasAss = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_ASSUMPTION", "Relation pointing to an input", 1.0);
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
		for(int i=0;i<ass.length;i++){
			new Relation(this, hasAss, ass[i]);
		}
	}
	
	/**
		Method for getting the assumptions for the current problem solving method.
		@return Assumption The assumptions related to the psm.
	*/
	public Assumption[] getAssumptions() throws Exception{
		Assumption[] ass = null;
		
		//Get all relations pointing to actions
		Relation rels[] = this.getLocalRelations("HAS_ASSUMPTION");
	
		/*
			Only one set of assumptions contained by one Assumptions object is allowed, 
			if any assumptions are declared.
		*/
		if(rels==null){
			//Do not do anything, return the null pointer already assigned.
		}
		else{ 
			for(int i=0; i<rels.length; i++){
				ass[i] = (Assumption)rels[i].getValue();	
			}
		}
		return ass;
	}
	
	/**
		Method for checking whether an action is activated or not.
		@return boolean The method returns true if activated and false if not.
	 */
	public boolean isActivated(){
		try{
			if(this.getRelation("HAS_ACTIVATION")!=null){
				return true;
			}
		}catch(NoSuchRelationTypeException e){
			//If there is no such relationtype, this task is not activated.
			//Nothing need to be done and false is returned.
		}catch(NoSuchRelationException e){
			//If there is no such relation, this task is not activated.
			//Nothing need to be done and false is returned.
		}
		return false;
	}
	
	/**
		Method for setting an action as activated.
	 */
	public void setAsActivated(){
		TMST tmst = (TMST)this.getKnowledgeModel();
		RelationType isActivated = tmst.getRelationType("HAS_ACTIVATION");
		
		if(isActivated==null){
			//Has to make new Relation type
			try{
				isActivated = new RelationType((KnowledgeModel)tmst, tmst.getRelationType(BasicModel.HAS_VALUE_TYPE), "HAS_ACTIVATION", "Relation pointing to an activated entity.", 1.0);
				//System.out.println(" -- Making new relation type: HAS_ACTIVATION");
			}catch(Exception e){
				e.printStackTrace();
			}
		}
		
		if(!tmst.hasEntity("ACTIVATED")){
			try{
				Entity activated = new Entity((KnowledgeModel)tmst, "ACTIVATED", "Entity for indicating whether a task, a psm or an action is activated.");
				activated.addRelation(BasicModel.INSTANCE_OF, tmst.getEntity(BasicModel.SYMBOL));	
				//System.out.println(" -- Making new entity: ACTIVATED");
			}catch(NameAlreadyExistException e){
				//Do nothing, it the entity already exist, there is no problem.
			}catch(NoSuchRelationTypeException e){
				//This should never happen.
				e.printStackTrace();
			}
		}
		new Relation(this, isActivated, tmst.getEntity("ACTIVATED"));
	}
	
	/**
		Method which has to be overriden (from the class EntityType)
		@param value The entity to match against.
		@return boolean The method returns true if the entity value matches the this entity type, and false if not.
	*/
	public static boolean matches(Entity value){
		TMST tmst = (TMST)value.getKnowledgeModel();
		Entity ent = tmst.getEntity("PROBLEM_SOLVING_METHOD");
		
		if (ent == null) return false; 
		
		Collection subclasses = ent.getAllSubclassesCollection();
		
		if(value.getName().equals("PROBLEM_SOLVING_METHOD")){
			return false;
		}else if(subclasses.contains(value)){
			return true;
		}
		return false;
	}
}